//////////////////////////////////////////////////////////////////////
// aft_bri_logic_channels.cpp - Device Manager Property Page for
//			A500 ISDN BRI adapter.
//
// Author	:	David Rokhvarg
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "sdlacfg.h"

#define front_end_cfg wandev_conf.fe_cfg

////////////////////////////////////////////////////////////////////////////////
//AFT cfg functions prototypes

static int AFT_OnNotify(HWND    ParentHwnd,
				LPNMHDR NmHdr,
				SETUPSTUFF * Params
				);

static void AFT_BRI_Init(HKEY	hkey, HWND hdlg, GLOBALS* cfg);

static int display_logic_channel_properties(int logic_ch_number, HWND hdlg);

static int selected_ind=0;
static LRESULT CALLBACK bri_logic_ch_properties_dlg(HWND hdlg, UINT iMessage,
										 WPARAM wParam, LPARAM lParam);

static int get_logic_channel_properties_data(HWND hdlg);
static BOOL aft_check_all_cfg(HWND hdlg);

static int DisplayContextMenu(HWND hdlg, POINT pt);
static HMENU hmenu;   // group menu 

static int get_number_of_channels(unsigned int chan_bit_map);
static int get_recommended_bitstream_mtu(int num_of_channels);

///////////////////////////////////////////////////////////////////////////////

static char tmp_char_buffer[TMP_BUFFER_LEN];

static char* aft_line_mode_options[] = {MODE_OPTION_HDLC, MODE_OPTION_BITSTRM, ISDN_BRI_DCHAN, NULL};

//options for "non BRI D-chan" modes
static char* aft_operational_mode_options[] = 
{	SDLA_DECODE_USEDBY_FIELD(API),
	SDLA_DECODE_USEDBY_FIELD(TDM_SPAN_VOICE_API),
	SDLA_DECODE_USEDBY_FIELD(TDM_CHAN_VOICE_API),
	SDLA_DECODE_USEDBY_FIELD(STACK),//another driver (sprotocol.sys or wanpipe.sys) will run above. 
	SDLA_DECODE_USEDBY_FIELD(TDM_VOICE_API),
	SDLA_DECODE_USEDBY_FIELD(API_LEGACY),
	NULL
};

//options for bri dchan mode
static char* aft_bri_dchan_operational_mode_options[] = 
{	SDLA_DECODE_USEDBY_FIELD(API),
	SDLA_DECODE_USEDBY_FIELD(API_LEGACY),
	SDLA_DECODE_USEDBY_FIELD(STACK),//another driver (sprotocol.sys or wanpipe.sys) will run above. 
	NULL
};

//these are recommendations only:
#define MIN_RECOMMENDED_BITSTREAM_MTU	256
#define MIN_RECOMMENDED_HDLC_MTU		2048

//D-chan MUST always be at 3
#define BRI_DCHAN_ACTIVE_CHAN	3

const look_up_t bri_clock_mode[] = {
	{"Normal (Default)",	WAN_NORMAL_CLK},
	{"Master",				WAN_MASTER_CLK},
	{NULL,					0}
};

///////////////////////////////////////////////////////////////////////////////
DLGPROC_RETURN_TYPE WINAPI AFT_BRI_PageDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{							// PageDlgProc
	int i;
	static char local_tmp_buff[TMP_BUFFER_LEN];

	switch (msg)
	{						// process message
	case WM_NOTIFY:
		Debug(DBG_AFT_BRI_PAGE, "WM_NOTIFY");
		//If there is a problem with the new configuration then refuse to close the property page and return TRUE.
		//If all checks are OK then allow the closing.
        if(AFT_OnNotify(hdlg, (NMHDR *)lParam, stuff) == HANDLE_ERROR)
		{	SetWindowLong(hdlg, DWL_MSGRESULT, PSNRET_INVALID);
			return TRUE;
		}
		break;

	case WM_INITDIALOG:
		// Load the menu resource. 
		if ((hmenu = LoadMenu(hInst, MAKEINTRESOURCE(IDR_MENU_IF_OPTIONS))) == NULL) 
		{	Debug(DBG_AFT_BRI_PAGE, "hmenu == NULL!!\n");
		}

		if (okay == TRUE)
		{	//initialize visual controls
			AFT_BRI_Init(hkey, hdlg, &Globals_New_Cfg);
		}

		if (!okay)
		{	// no SampleInfo
			ShowWindow(GetDlgItem(hdlg, IDC_INFOLABEL), SW_HIDE);
			ShowWindow(GetDlgItem(hdlg, IDB_MOREINFO), SW_HIDE);
		}
		break;
		
   	case WM_DESTROY:
		Debug(DBG_AFT_BRI_PAGE, "Samclass:WM_DESTROY");
		{
			LPSTR infofile = (LPSTR) GetWindowLongPtr(hdlg, DWL_USER);
			if (infofile)
			{	GlobalFree((HGLOBAL) infofile);
			}

			if(hmenu)
			{	// Destroy the menu. 
				DestroyMenu(hmenu); 
			}
		}
		break;
		
	case WM_SHOWWINDOW:	
		//This message received each time the Dialog shown/hidden.
		//Update visual controls with latest information stored in "Globals_New_Cfg".
		Debug(DBG_AFT_BRI_PAGE, "DBG_AFT_BRI_PAGE: WM_SHOWWINDOW");
		set_dialog_wanpipe_name(hdlg);
		AFT_BRI_Init(hkey, hdlg, &Globals_New_Cfg);
		break;

	case WM_CONTEXTMENU:
		Debug(DBG_AFT_BRI_PAGE, "WM_CONTEXTMENU\n");
		//make sure:
		//1. mouse above LIST_GROUP_PROPERTIES
		//
		//2. there is an active selection in LIST_GROUP_PROPERTIES
		//   before displaying the menu.
		{ 
			RECT	rc;					// client area of window 
			int		x = GET_X_LPARAM(lParam);
			int		y = GET_Y_LPARAM(lParam);
			POINT	pt = { x, y };      // location of mouse click 
			int		nBuffer[BIGBUFF];
			int		nSelItems;
#if 0
			int		nSelItemsInBuffer;
#endif
			// Get the bounding rectangle of the client area. 
			GetClientRect(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), &rc); 
 
			// Convert the mouse position to client coordinates. 
			ScreenToClient(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), &pt); 
				
			// If the position is in the client area, display a  
			// shortcut menu. 
			if (PtInRect(&rc, pt)) 
			{
#if 1
				//this is single-selection version
				selected_ind = (int)SendDlgItemMessage(hdlg, LIST_GROUP_PROPERTIES, LB_GETCURSEL, 0, 0);
				if(selected_ind == LB_ERR)
				{	MessageBox(GetActiveWindow(), "No Group was selected", "", MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
					break;
				}
				nSelItems = 1;
#else				
				//this is the multiple-selection version
				nSelItems = SendMessage(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), LB_GETSELCOUNT, 0, 0); 
				Debug(DBG_AFT_BRI_PAGE, "nSelItems: %d\n", nSelItems);
				if(nSelItems == 0)
				{	MessageBox(GetActiveWindow(), "No Group was selected. Make sure at least one is selected.", 
						"sdlacfg",	MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
					break;
				}
						
				nSelItemsInBuffer = 
					SendMessage(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), LB_GETSELITEMS, BIGBUFF, (LPARAM) nBuffer);
				Debug(DBG_AFT_BRI_PAGE, "nSelItemsInBuffer: %d\n", nSelItemsInBuffer);

				if (nSelItems > nSelItemsInBuffer)
				{ 	MessageBox(hdlg, "Too many items selected.", NULL, MB_OK); 
					break;
				}

				//use the 1-st SELECTED line as 'selected_ind'.
				selected_ind = nBuffer[0];
#endif
				ClientToScreen(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), &pt); 
				switch(DisplayContextMenu(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), pt))
				{
				case MENU_ITEM_MONITOR:
					if(is_legacy_api_mode(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode) == false){
						//Interface name e.g.: WANPIPE1_IF1 - if names start from 1
						start_monitor(selected_ind + 1);
					}else{
						start_monitor(selected_ind);
					}
					break;

				case MENU_ITEM_IF_PROPERTIES:
					display_logic_channel_properties(selected_ind, hdlg);

					if(nSelItems > 1)
					{	//more than one group selected.
						int first_active_chan = atoi(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].active_ch_str);
						int chan_counter = 1;
						Debug(DBG_AFT_BRI_PAGE, "first_active_chan: %d\n", first_active_chan);

						//copy the same configuration to all seleted itmes.
						for (i = 1; i < nSelItems; i++) 
						{	Debug(DBG_AFT_BRI_PAGE, "nBuffer[%d]: %d\n", i, nBuffer[i]);

							memcpy(	&Globals_New_Cfg.aft_logic_ch_cfg[nBuffer[i]],
									&Globals_New_Cfg.aft_logic_ch_cfg[selected_ind],
									sizeof(AFT_LOGIC_CH_CFG));

							//set active channels sequentially
							_snprintf(	Globals_New_Cfg.aft_logic_ch_cfg[nBuffer[i]].active_ch_str,
										TMP_BUFFER_LEN,
										"%d",
										first_active_chan+chan_counter);
										//(1 << (first_active_chan+chan_counter)));
							chan_counter++;
						} 
					}

					//re-initialize visual controls in case was changed
					AFT_BRI_Init(hkey, hdlg, &Globals_New_Cfg);
					break;
				}//switch()
			}//if() 
		} 
		break;

    case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case BUTTON_HELP:
			WinHelp(GetActiveWindow(), "WanpipePnp.hlp", HELP_CONTEXT, IDH_LINE_CFG);
			break;

		case COMBO_NUMBER_OF_GROUPS:
			if (GET_WM_COMMAND_CMD(wParam, lParam) == CBN_SELCHANGE)
			{	//Number of groups changed, ask user to confirm, cause it is a serious change.
				if(yes_no_question("Are you sure you want change the number of groups?") == IDYES)
				{	i = SendDlgItemMessage(hdlg, COMBO_NUMBER_OF_GROUPS, CB_GETCURSEL, 0, 0);
					SendDlgItemMessage(	hdlg, COMBO_NUMBER_OF_GROUPS, CB_GETLBTEXT, i, (LPARAM)local_tmp_buff);
					Globals_New_Cfg.aft_number_logic_channels = (unsigned short)atoi(local_tmp_buff);
				}
				//reinitialize visual controls with updated cfg
				AFT_BRI_Init(hkey, hdlg, &Globals_New_Cfg);
			}
			break;
		}
		return TRUE;
	}
	return FALSE;
}

///////////////////////////////////////////////////////////////////////////////

int DisplayContextMenu(HWND hdlg, POINT pt) 
{ 
	HMENU hmenuTrackPopup;  // shortcut menu 
	int menu_item_id;

	if(!hmenu){
		return 0;
	}

	// TrackPopupMenu cannot display the menu bar so get 
	// a handle to the first shortcut menu. 
	hmenuTrackPopup = GetSubMenu(hmenu, 0); 
 
	// Display the shortcut menu. Track both mouse buttons. 
	menu_item_id = TrackPopupMenu(	hmenuTrackPopup, 
							TPM_LEFTALIGN | TPM_RIGHTBUTTON | TPM_RETURNCMD, 
							pt.x,
							pt.y,
							0,
							hdlg, NULL); 

	Debug(DBG_AFT_BRI_PAGE, "menu_item_id: %d\n", menu_item_id);
	return menu_item_id;
} 

///////////////////////////////////////////////////////////////////////////////

UINT CALLBACK AFT_BRI_PageCallbackProc(HWND junk, UINT msg, LPPROPSHEETPAGE p)
{
	if (msg == PSPCB_RELEASE && p->lParam)
	{
		Debug(DBG_ANALOG_PAGE, "PSPCB_RELEASE : %s()", __FUNCTION__);
		RegCloseKey(hkey);

		//compare the configuration structures, restart if different.
		if((released_by_OK_button == 1) &&  memcmp(&Globals_Old_Cfg, &Globals_New_Cfg, sizeof(Globals_New_Cfg)))
		{
			if(yes_no_question("Device Configuration was changed. Do you want the Device to be restarted now?")
					== IDYES){
				restart_device();
			}else{
				MessageBox(GetActiveWindow(), (LPCTSTR)"Device was NOT restarted.",
					"Note for the user", MB_OK | MB_ICONINFORMATION | MB_SYSTEMMODAL );
			}
		}
		delete (SETUPSTUFF*) p->lParam;
		return 0;//return code is ignored
	}
	return TRUE;	// matters only for PSPCB_CREATE
}

///////////////////////////////////////////////////////////////////////////////

static int AFT_OnNotify(	HWND		ParentHwnd,
							LPNMHDR		NmHdr,
							SETUPSTUFF	*Params	)
{
    SP_DEVINSTALL_PARAMS	spDevInstall = {0};
    TCHAR					friendlyName[LINE_LEN] ={0};
	BOOL					configuration_check_passed = FALSE;
    
    switch (NmHdr->code) 
	{
    case PSN_APPLY:	// Sent when the user clicks on Apply OR OK !!
		Debug(DBG_AFT_BRI_PAGE, "AFT_OnNotify((): PSN_APPLY");

		configuration_check_passed = aft_check_all_cfg(ParentHwnd);
		
		released_by_OK_button = 1;

		if(configuration_check_passed == TRUE){
			return HANDLE_OK;
		}else{
			return HANDLE_ERROR;
		}
		break;

	case PSN_QUERYCANCEL:
		Debug(DBG_AFT_BRI_PAGE, "AFT_OnNotify(() : PSN_QUERYCANCEL");
		released_by_OK_button = 0;
    default:
		Debug(DBG_AFT_BRI_PAGE, "AFT_OnNotify(() : default case:");
        return HANDLE_DEFAULT;
    }
    return HANDLE_DEFAULT;   
}

///////////////////////////////////////////////////////////////////////////////

static BOOL aft_check_all_cfg(HWND hdlg)
{	
	int i;
	unsigned long tmp_active_ch = 0;
	unsigned long cumulutive_active_ch = 0;
	char local_tmp_buff[TMP_BUFFER_LEN];
	char local_tmp_buff_1[TMP_BUFFER_LEN];

	Debug(DBG_AFT_BRI_PAGE, "%s()", __FUNCTION__);
	
	//get number of logic channels
	i = SendDlgItemMessage(hdlg, COMBO_NUMBER_OF_GROUPS, CB_GETCURSEL, 0, 0);
	SendDlgItemMessage(	hdlg, COMBO_NUMBER_OF_GROUPS, CB_GETLBTEXT, i, (LPARAM)local_tmp_buff);
	Globals_New_Cfg.aft_number_logic_channels = (unsigned short)atoi(local_tmp_buff);
	
	for(i = 0; i < Globals_New_Cfg.aft_number_logic_channels; i++){
		//check all logic channels were configured
		if(!strcmp(Globals_New_Cfg.aft_logic_ch_cfg[i].active_ch_str, NOT_INIT_STR)){
			_snprintf(local_tmp_buff, TMP_BUFFER_LEN, "Group %d - no BRI channels specified!", i + 1);
			MessageBox(GetActiveWindow(), local_tmp_buff, "Configuration Error", MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
			return FALSE;
		}

		//check the channels string
		if(strcmp(Globals_New_Cfg.aft_logic_ch_cfg[i].line_mode, ISDN_BRI_DCHAN) == 0){
			tmp_active_ch = 1 << BRI_DCHAN_ACTIVE_CHAN;//0x4
		}else{
			tmp_active_ch =	parse_active_channel(Globals_New_Cfg.aft_logic_ch_cfg[i].active_ch_str,
				Globals_New_Cfg.front_end_cfg.media,
				FALSE);
			if(tmp_active_ch == 0){
				_snprintf(local_tmp_buff, TMP_BUFFER_LEN, "Group %d - Error found in \"BRI channels\" string!", i + 1);
				MessageBox(GetActiveWindow(), local_tmp_buff, "Configuration Error", MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
				return FALSE;
			}
		}

		//check channels are NOT in use already
		if(tmp_active_ch & cumulutive_active_ch){
			_snprintf(local_tmp_buff, TMP_BUFFER_LEN,
				"Group %d - one or more \"BRI channels\" already used by another group!", i + 1);
			MessageBox(GetActiveWindow(), local_tmp_buff, "Configuration Error", MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
			return FALSE;
		}
		//update cumulative channels
		cumulutive_active_ch |= tmp_active_ch;
	}

	Debug(DBG_AFT_BRI_PAGE, "*********  Writing AFT Configuration  ********\n");

	//the check passed, save to registry
	SetRegistryStringValue(hkey, "aft_number_of_logic_channels", Globals_New_Cfg.aft_number_logic_channels);

	for(i=0; i < Globals_New_Cfg.aft_number_logic_channels; i++)
	{	//active channels
		snprintf(local_tmp_buff, TMP_BUFFER_LEN, "aft_logic_channel_%d_active_ch", i);
		SetRegistryStringValue(hkey,  local_tmp_buff,
			(char*)Globals_New_Cfg.aft_logic_ch_cfg[i].active_ch_str);

		//line mode
		snprintf(local_tmp_buff, TMP_BUFFER_LEN, "aft_logic_channel_%d_line_mode", i);
		SetRegistryStringValue(hkey, local_tmp_buff,
			(char*)Globals_New_Cfg.aft_logic_ch_cfg[i].line_mode);

		//MTU
		snprintf(local_tmp_buff, TMP_BUFFER_LEN, "aft_logic_channel_%d_mtu", i);
		SetRegistryStringValue(hkey, local_tmp_buff,
			(char*)Globals_New_Cfg.aft_logic_ch_cfg[i].mtu);

		//operation mode
		snprintf(local_tmp_buff, TMP_BUFFER_LEN, "aft_logic_channel_%d_operational_mode", i);
		SetRegistryStringValue(hkey, local_tmp_buff,
			(char*)Globals_New_Cfg.aft_logic_ch_cfg[i].operational_mode);

		//Idle char. Write in DECIMAL !!!
		snprintf(local_tmp_buff, TMP_BUFFER_LEN, "aft_logic_channel_%d_idle_char", i);
	
		_itoa(Globals_New_Cfg.aft_logic_ch_cfg[i].decimal_idle_char, local_tmp_buff_1, 10);

		SetRegistryStringValue(hkey, local_tmp_buff, local_tmp_buff_1);
	}

	///////////////////////////////////////////////////////////////
	i = SendDlgItemMessage(hdlg, COMBO_BRI_CLOCK_MODE, CB_GETCURSEL, 0, 0);
	if(CB_ERR != i){
		//set new value only if something is selected
		Globals_New_Cfg.front_end_cfg.cfg.bri.clock_mode = bri_clock_mode[i].value;
		SetRegistryStringValue(hkey, "aft_bri_clock_mode", Globals_New_Cfg.front_end_cfg.cfg.bri.clock_mode);
	}
	Debug(DBG_AFT_BRI_PAGE, "*********  Finished writing AFT Configuration  ********\n");
	return TRUE;
}
///////////////////////////////////////////////////////////////////////////////

static void AFT_BRI_Init(HKEY	hkey, HWND hdlg, GLOBALS* cfg)
{
	int i, current_select_index;
	char local_tmp_buff[TMP_BUFFER_LEN];

	//////////////////////////////////////////////////////////////////////
	SetDlgItemText(hdlg, BRI_MODULE_TYPE,
		//"hello"
		Globals_New_Cfg.BriModuleType
		);

	//////////////////////////////////////////////////////////////////////////////////////
	//number of groups may change so remove all and redraw.
	SendDlgItemMessage(hdlg, COMBO_NUMBER_OF_GROUPS, CB_RESETCONTENT, 0, 0);

	//Add strings to Number of groups combobox.
	for(i = 0; i < MAX_BRI_TIMESLOTS; i++){
		snprintf(local_tmp_buff, TMP_BUFFER_LEN, "%d", i + 1);
		SendDlgItemMessage(hdlg, COMBO_NUMBER_OF_GROUPS, CB_ADDSTRING, 0, (LPARAM)local_tmp_buff);
	}
	Debug(DBG_AFT_BRI_PAGE, "%s(): aft_number_logic_channels: %i\n", __FUNCTION__, cfg->aft_number_logic_channels);

	//make the current one selected
	SendDlgItemMessage(hdlg, COMBO_NUMBER_OF_GROUPS, CB_SETCURSEL, cfg->aft_number_logic_channels - 1, 0);

	//////////////////////////////////////////////////////////////////////////////////////
	//clear before adding
	SendMessage(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), LB_RESETCONTENT, 0, 0); 

	for(i = 0; i < cfg->aft_number_logic_channels; i++){
		AFT_LOGIC_CH_CFG *aft_logic_ch_cfg_ptr;

		if(is_legacy_api_mode(Globals_New_Cfg.aft_logic_ch_cfg[i].operational_mode) == false){
			//Interface name e.g.: WANPIPE1_IF1 - if names start from 1
			_snprintf(local_tmp_buff, TMP_BUFFER_LEN, "%s_IF%d\t", Globals_New_Cfg.DeviceName, i + 1);
		}else{
			//Interface name e.g.: WANPIPE1_IF0 - if names start from 0
			_snprintf(local_tmp_buff, TMP_BUFFER_LEN, "%s_IF%d\t", Globals_New_Cfg.DeviceName, i);
		}

		aft_logic_ch_cfg_ptr = &cfg->aft_logic_ch_cfg[i];

		//channels used by the Interface e.g.: 1-24
		_snprintf(&local_tmp_buff[strlen(local_tmp_buff)],
			TMP_BUFFER_LEN-strlen(local_tmp_buff),
			"%5s  ",//"channels: %s,\t",
			aft_logic_ch_cfg_ptr->active_ch_str);

		//line mode (HDLC / BitStream) used by the group
		_snprintf(&local_tmp_buff[strlen(local_tmp_buff)],
			TMP_BUFFER_LEN-strlen(local_tmp_buff),
			"%15s\t     ",//"line mode: %s,\t",
			aft_logic_ch_cfg_ptr->line_mode);

		//MTU used by the group
		_snprintf(&local_tmp_buff[strlen(local_tmp_buff)],
			TMP_BUFFER_LEN-strlen(local_tmp_buff),
			"%4s\t    ",//"MTU: %s\t",
			aft_logic_ch_cfg_ptr->mtu);

		//operational mode (API / STACK/ Router/ TDM VOICE API) used by the group
		_snprintf(&local_tmp_buff[strlen(local_tmp_buff)],
				TMP_BUFFER_LEN-strlen(local_tmp_buff),
				"%13s",
				aft_logic_ch_cfg_ptr->operational_mode);
		
		//display it all
		SendMessage(GetDlgItem(hdlg, LIST_GROUP_PROPERTIES), LB_ADDSTRING, 0, (LPARAM)local_tmp_buff); 
	}

	///////////////////////////////////////////////////////////////
	SendDlgItemMessage(hdlg, COMBO_BRI_CLOCK_MODE, CB_RESETCONTENT, 0, 0);
	i = 0; current_select_index = 0;
	while(bri_clock_mode[i].name != NULL){
		SendDlgItemMessage(hdlg, COMBO_BRI_CLOCK_MODE, CB_ADDSTRING, 0, (LPARAM)bri_clock_mode[i].name);
		if(bri_clock_mode[i].value == Globals_New_Cfg.front_end_cfg.cfg.bri.clock_mode){
			current_select_index = i;
		}
		i++;
	}
	SendDlgItemMessage(hdlg, COMBO_BRI_CLOCK_MODE, CB_SETCURSEL, current_select_index, 0);	


}

///////////////////////////////////////////////////////////////////////////////

static int display_logic_channel_properties(int logic_ch_number, HWND hdlg)
{
	INT_PTR rval;

	rval = DialogBox(hInst, MAKEINTRESOURCE(AFT_BRI_LOGIC_CHAN_PROPERTIES), hdlg, (DLGPROC)bri_logic_ch_properties_dlg);

	if(rval == -1){
		LPVOID lpMsgBuf;
		FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			FORMAT_MESSAGE_FROM_SYSTEM | 
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			GetLastError(),
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
			(LPTSTR) &lpMsgBuf,
			0,
			NULL 
		);
		return FALSE;
	}
	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////

static void disable_enable_bri_dchan_controls(HWND hdlg, BOOL flag)
{
	Debug(DBG_AFT_BRI_PAGE, "%s(): flag: %i\n", __FUNCTION__, flag);
	EnableWindow( GetDlgItem(hdlg, EDIT_ACTIVE_CHANNELS),		flag);
	EnableWindow( GetDlgItem(hdlg, EDIT_MTU),					flag);
}

///////////////////////////////////////////////////////////////////////////////

static void init_bri_logic_ch_properties(HWND hdlg)
{
	static char local_tmp_buff[TMP_BUFFER_LEN];
	int i, current_mode_index;

	Debug(DBG_AFT_BRI_PAGE, "%s()\n", __FUNCTION__);
	//////////////////////////////////////////////////////////////////////
	//set dialog's title		
	_snprintf(local_tmp_buff, TMP_BUFFER_LEN, "Properties of: %s_IF%d", Globals_New_Cfg.DeviceName, selected_ind);
	SetWindowText(hdlg, local_tmp_buff);

	///////////////////////////////////////////////////////////////////////
	//set the channels string
	SetDlgItemText(hdlg, EDIT_ACTIVE_CHANNELS, Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].active_ch_str);

	///////////////////////////////////////////////////////////////////////
	//add strings to Line Mode combobox
	SendDlgItemMessage(hdlg, COMBO_LINE_MODE, CB_RESETCONTENT, 0, 0);
	i = 0;
	while(aft_line_mode_options[i] != NULL){
		SendDlgItemMessage(hdlg, COMBO_LINE_MODE, CB_ADDSTRING, 0, (LPARAM)aft_line_mode_options[i]);
		i++;
	}

	//make the current one selected
	i = SendDlgItemMessage(	hdlg, COMBO_LINE_MODE, CB_FINDSTRING, -1,
							(LPARAM)Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode);

	SendDlgItemMessage(hdlg, COMBO_LINE_MODE, CB_SETCURSEL, (i != CB_ERR ? i : 0), 0);

	Debug(DBG_AFT_BRI_PAGE, "line_mode: %s, selected_ind: %i\n", 
		Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, selected_ind);

	if(strcmp(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, ISDN_BRI_DCHAN) == 0){
		disable_enable_bri_dchan_controls(hdlg, FALSE);
	}else{
		disable_enable_bri_dchan_controls(hdlg, TRUE);
	}

	///////////////////////////////////////////////////////////////////////
	//set the MTU value
	SetDlgItemText(hdlg, EDIT_MTU, Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].mtu);

	///////////////////////////////////////////////////////////////////////
	//add strings to Operational Mode combobox
	Debug(DBG_AFT_BRI_PAGE, "Globals_New_Cfg.aft_logic_ch_cfg[%d].operational_mode: %s\n", 
		selected_ind, Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode);

	Debug(DBG_AFT_BRI_PAGE, "Globals_New_Cfg.aft_logic_ch_cfg[%d].line_mode: %s\n", 
		selected_ind, Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode);

	SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_RESETCONTENT, 0, 0);
	i = 0;
	if(strcmp(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, ISDN_BRI_DCHAN) == 0){
		while(aft_bri_dchan_operational_mode_options[i] != NULL){
			SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_ADDSTRING, 0, (LPARAM)aft_bri_dchan_operational_mode_options[i]);
			if(strcmp(aft_bri_dchan_operational_mode_options[i], Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode) == 0){
				current_mode_index = i;
			}
			i++;
		}
	}else{
		while(aft_operational_mode_options[i] != NULL){
			SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_ADDSTRING, 0, (LPARAM)aft_operational_mode_options[i]);
			if(strcmp(aft_operational_mode_options[i], Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode) == 0){
				current_mode_index = i;
			}
			i++;
		}
	}

	Debug(DBG_AFT_BRI_PAGE, "current_mode_index: %d\n", current_mode_index);

	if(current_mode_index > SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_GETCOUNT, 0, 0) - 1){
		Debug(DBG_AFT_BRI_PAGE, "current_mode_index: %d is outside of valid range of: %d\n",
			current_mode_index, SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_GETCOUNT, 0, 0) - 1);
		current_mode_index = 0;
	}
	//make the current one selected
	SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_SETCURSEL, current_mode_index, 0);

	///////////////////////////////////////////////////////////////////////
	//Show Idle char value in HEX!!!
	snprintf(local_tmp_buff, TMP_BUFFER_LEN, "%02X", Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].decimal_idle_char);
	SetDlgItemText(hdlg, EDIT_IDLE_CHAR, local_tmp_buff);
}

///////////////////////////////////////////////////////////////////////////////

static LRESULT CALLBACK bri_logic_ch_properties_dlg(HWND hdlg, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	int i;
	static char local_tmp_buff[TMP_BUFFER_LEN];

	switch(iMessage) {
	case WM_SYSCOMMAND:
		switch(wParam & 0xfff0) {
		case SC_CLOSE:	/* Don't allow user to close window this way */
			return(1);
		}
		break;

	case WM_INITDIALOG:
		Debug(DBG_AFT_BRI_PAGE, "bri_logic_ch_properties_dlg(): WM_INITDIALOG\n");
		init_bri_logic_ch_properties(hdlg);
		return(TRUE);

	case WM_COMMAND:
		switch(GET_WM_COMMAND_ID(wParam, lParam)) 
		{
		case IDOK:
			if(get_logic_channel_properties_data(hdlg) == TRUE){
				EndDialog(hdlg, 0);
			}
			break;

		case IDCANCEL:
			EndDialog(hdlg, 0);
			//user cancelled any changes, restore original values for the group
			memcpy(	&Globals_New_Cfg.aft_logic_ch_cfg[selected_ind],
					&Globals_Old_Cfg.aft_logic_ch_cfg[selected_ind],
					sizeof(AFT_LOGIC_CH_CFG));
			break;

		case COMBO_LINE_MODE:
			Debug(DBG_AFT_BRI_PAGE, "action on COMBO_LINE_MODE\n");

			if (GET_WM_COMMAND_CMD(wParam, lParam) != CBN_SELCHANGE){
				//no change - nothing to do
				break;
			}

			if((i = SendDlgItemMessage(hdlg, COMBO_LINE_MODE, CB_GETCURSEL, 0, 0)) == CB_ERR){
				break;
			}
			SendDlgItemMessage(	hdlg, COMBO_LINE_MODE, CB_GETLBTEXT, i, (LPARAM)local_tmp_buff);
			sprintf(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, "%s", local_tmp_buff);

			Debug(DBG_AFT_BRI_PAGE, "local_tmp_buff: %s\n", local_tmp_buff);

			if(strcmp(local_tmp_buff, MODE_OPTION_BITSTRM) == 0){

				if((i = SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_GETCURSEL, 0, 0)) == CB_ERR){
					break;
				}
				SendDlgItemMessage(	hdlg, COMBO_OPERATIONAL_MODE, CB_GETLBTEXT, i, (LPARAM)local_tmp_buff);

				if(	strcmp(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode, 
														SDLA_DECODE_USEDBY_FIELD(STACK)) == 0 ||
					strcmp(local_tmp_buff, 				SDLA_DECODE_USEDBY_FIELD(STACK)) == 0 ){

					//router should *not* run in BitStream mode, so set it to API
					strcpy(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode, SDLA_DECODE_USEDBY_FIELD(API));
					
					//make current mode selected
					if((i = SendDlgItemMessage(	hdlg, COMBO_OPERATIONAL_MODE, CB_FINDSTRING, -1,
						(LPARAM)Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode)) == CB_ERR){
						break;
					}
					SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_SETCURSEL, i, 0);
				}
				SendMessage(hdlg, WM_INITDIALOG, 0, 0);
			}else if(strcmp(local_tmp_buff, MODE_OPTION_HDLC) == 0 || strcmp(local_tmp_buff, ISDN_BRI_DCHAN) == 0){

				if(strcmp(local_tmp_buff, ISDN_BRI_DCHAN) == 0){
					sprintf(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].active_ch_str, "%i", BRI_DCHAN_ACTIVE_CHAN);
					sprintf(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].mtu, "%i", MAX_DFRAME_LEN_L1);
				}

				if((i = SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_GETCURSEL, 0, 0)) == CB_ERR){
					break;
				}
				SendDlgItemMessage(	hdlg, COMBO_OPERATIONAL_MODE, CB_GETLBTEXT, i, (LPARAM)local_tmp_buff);
				
				if( strcmp(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode,
							SDLA_DECODE_USEDBY_FIELD(TDM_VOICE_API)) == 0	||
					strcmp(local_tmp_buff, 		
							SDLA_DECODE_USEDBY_FIELD(TDM_VOICE_API)) == 0	||
					strcmp(local_tmp_buff, 		
							SDLA_DECODE_USEDBY_FIELD(TDM_CHAN_VOICE_API)) == 0	||
					strcmp(local_tmp_buff, 		
							SDLA_DECODE_USEDBY_FIELD(TDM_SPAN_VOICE_API)) == 0	){

					//Voice API is INVALID for HDLC, so set it to API
					strcpy(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode, 
							SDLA_DECODE_USEDBY_FIELD(API));
					
					//make current mode selected
					if((i = SendDlgItemMessage(	hdlg, COMBO_OPERATIONAL_MODE, CB_FINDSTRING, -1,
						(LPARAM)Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode)) == CB_ERR){
						break;
					}
					SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_SETCURSEL, i, 0);
				}
				SendMessage(hdlg, WM_INITDIALOG, 0, 0);
			}
			break;

		case COMBO_OPERATIONAL_MODE:
			Debug(DBG_AFT_BRI_PAGE, "action on COMBO_OPERATIONAL_MODE\n");
			if (GET_WM_COMMAND_CMD(wParam, lParam) != CBN_SELCHANGE){
				//no change - nothing to do
				break;
			}
			if((i = SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_GETCURSEL, 0, 0)) == CB_ERR){
				break;
			}
			SendDlgItemMessage(	hdlg, COMBO_OPERATIONAL_MODE, CB_GETLBTEXT, i, (LPARAM)local_tmp_buff);

			///////////////
			strcpy(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode, local_tmp_buff);
			///////////////

			if(	strcmp(local_tmp_buff, SDLA_DECODE_USEDBY_FIELD(TDM_VOICE_API)) == 0	||
				strcmp(local_tmp_buff, SDLA_DECODE_USEDBY_FIELD(TDM_CHAN_VOICE_API)) == 0 ||
				strcmp(local_tmp_buff, SDLA_DECODE_USEDBY_FIELD(TDM_SPAN_VOICE_API)) == 0	){

				//HDLC is invalid for Voice API, so set it to BitStream
				strcpy(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, MODE_OPTION_BITSTRM);
					
				//make current mode selected
				if((i = SendDlgItemMessage(	hdlg, COMBO_LINE_MODE, CB_FINDSTRING, -1,
					(LPARAM)Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode)) == CB_ERR){
					break;
				}
				SendDlgItemMessage(hdlg, COMBO_LINE_MODE, CB_SETCURSEL, i, 0);

			}else if(strcmp(local_tmp_buff, SDLA_DECODE_USEDBY_FIELD(STACK)) == 0){

				if(strcmp(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, ISDN_BRI_DCHAN) != 0){
					//BitStream is invalid for STACK, so set it to HDLC
					strcpy(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, MODE_OPTION_HDLC);
				}

				//make current mode selected
				if((i = SendDlgItemMessage(	hdlg, COMBO_LINE_MODE, CB_FINDSTRING, -1,
					(LPARAM)Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode)) == CB_ERR){
					break;
				}
				SendDlgItemMessage(hdlg, COMBO_LINE_MODE, CB_SETCURSEL, i, 0);
			}
			SendMessage(hdlg, WM_INITDIALOG, 0, 0);
			break;
		}
	}
	return(0);
}

//////////////////////////////////////////////////////////////////////////////

static int get_logic_channel_properties_data(HWND hdlg)
{
	int		j, rc, mtu;
	char	local_tmp_buff[TMP_BUFFER_LEN];

	///////////////////////////////////////////////////////////////////////
	//get line mode
	j = SendDlgItemMessage(hdlg, COMBO_LINE_MODE, CB_GETCURSEL, 0, 0);
	strcpy(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, aft_line_mode_options[j]);

	///////////////////////////////////////////////////////////////////////
	//get the channels string
	//do not copy directly to 'New_Cfg', because in case the check fails,
	//the wrong string already copied!!
	GetDlgItemText(hdlg, EDIT_ACTIVE_CHANNELS, (char*)local_tmp_buff, TMP_BUFFER_LEN);

	if(strcmp(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, ISDN_BRI_DCHAN) == 0){
		sprintf(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].active_ch_str, "%i", BRI_DCHAN_ACTIVE_CHAN);
	}else{
		if(CheckActiveChannels(
			(unsigned char*)local_tmp_buff,
			//(unsigned char*)Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].active_ch_str,
			Globals_New_Cfg.front_end_cfg.media) == FALSE){

			MessageBox(GetActiveWindow(), 
				"Error found in \"BRI channels\" string.", "Configuration Error",
				MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
			return FALSE;
		}
	}

	//passed the check, copy to the 'NewCfg'
	strcpy(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].active_ch_str, local_tmp_buff);

	//get MTU
	GetDlgItemText(hdlg, EDIT_MTU,
		Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].mtu,
		TMP_BUFFER_LEN);

	mtu = atoi(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].mtu);
	if(mtu < MIN_AFT_MTU || mtu > MAX_AFT_MTU){

		_snprintf(local_tmp_buff, TMP_BUFFER_LEN,
			"Invalid value in \"MTU\". Minimum is %d, Maximum is %d.",
			MIN_AFT_MTU, MAX_AFT_MTU);

		MessageBox(GetActiveWindow(), local_tmp_buff, "Configuration Error",															//szTitleString,
			MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
		return FALSE;
	}

	if(mtu % 4){
		_snprintf(local_tmp_buff, TMP_BUFFER_LEN,
			"Invalid value in \"MTU\". Must be divizible by 4.");

		MessageBox(GetActiveWindow(), local_tmp_buff, "Configuration Error",															//szTitleString,
			MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
		return FALSE;
	}

	//if BitStream, check MTU is also divisible by the number of channels (timeslots) in the group
	if(strcmp(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, MODE_OPTION_BITSTRM) == 0){

		int num_of_channels = 
			get_number_of_channels(
				parse_active_channel(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].active_ch_str,
										Globals_New_Cfg.front_end_cfg.media,
										FALSE));
	
		Debug(DBG_AFT_BRI_PAGE, "num_of_channels: %d\n", num_of_channels);
/*
		if(mtu % num_of_channels){
_snprintf(local_tmp_buff, TMP_BUFFER_LEN,
	"Invalid \"MTU\". For 'BitStream' Must be divisible by 4 AND by the number of channels in the group (%d). Recommended MTU: %d.",
	num_of_channels, get_recommended_bitstream_mtu(num_of_channels));

			MessageBox(GetActiveWindow(), local_tmp_buff, "Configuration Error",															//szTitleString,
				MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
			return FALSE;
		}
*/
	}else if(strcmp(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, MODE_OPTION_HDLC) == 0){

		if(mtu < MIN_RECOMMENDED_HDLC_MTU){

			_snprintf(local_tmp_buff, TMP_BUFFER_LEN,
				"The recommended \"MTU\" for HDLC is %d. Current \"MTU\" is %d. Would you like to use the recommended \"MTU\"?",
				MIN_RECOMMENDED_HDLC_MTU, mtu);

			if(yes_no_question(local_tmp_buff) == IDYES){
				//set the MTU value
				_snprintf(	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].mtu, TMP_BUFFER_LEN, "%d", 
							MIN_RECOMMENDED_HDLC_MTU);
				SetDlgItemText(hdlg, EDIT_MTU, Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].mtu);
			}
		}
	}

	//get operational mode
	j = SendDlgItemMessage(hdlg, COMBO_OPERATIONAL_MODE, CB_GETCURSEL, 0, 0);
	if(strcmp(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].line_mode, ISDN_BRI_DCHAN) == 0){
		strcpy(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode, aft_bri_dchan_operational_mode_options[j]);
	}else{
		strcpy(Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode, aft_operational_mode_options[j]);
	}
	Debug(DBG_AFT_BRI_PAGE, "%s(): Globals_New_Cfg.aft_logic_ch_cfg[%d].operational_mode: %s \n",
		__FUNCTION__, selected_ind, Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].operational_mode);

	//get Idle char
	GetDlgItemText(hdlg, EDIT_IDLE_CHAR, local_tmp_buff, TMP_BUFFER_LEN);

	rc = sscanf(local_tmp_buff, "%X", &j);
	if(rc == 0 || rc == EOF){
		_snprintf(local_tmp_buff, TMP_BUFFER_LEN,
			"Invalid value in \"Idle char\". Has to be a Hexadecimal char (i.g.: 7E).");

		MessageBox(GetActiveWindow(), local_tmp_buff, "Configuration Error", //szTitleString,
			MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
		return FALSE;
	}
	Globals_New_Cfg.aft_logic_ch_cfg[selected_ind].decimal_idle_char = (unsigned char)j;
	Debug(DBG_AFT_BRI_PAGE, "idle_char: 0x%02X (decimal: %d)\n", j, j);

	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////

static int get_number_of_channels(unsigned int chan_bit_map)
{
	int chan_counter = 0, i;

	for(i = 0; i < sizeof(unsigned int)*8; i++){
		if(chan_bit_map & (1 << i)){
			chan_counter++;
		}
	}
	return chan_counter;
}

static int get_recommended_bitstream_mtu(int num_of_channels)
{
	int recommended_mtu = num_of_channels * 32;//<--- will work well in most cases

	//
	//Smaller mtu will cause more interrupts.
	//Recommend at least MIN_RECOMMENDED_BITSTREAM_MTU, user will set to smaller if timing is important.
	//
	while(recommended_mtu < MIN_RECOMMENDED_BITSTREAM_MTU){
		recommended_mtu = recommended_mtu * 2;
	}

	return recommended_mtu;
}

//////////////////////////////////////////////////////////////////////////////
/*
1. fixed BRI D-channel transmission return code mismatch between Front End and AFT core code. This could cause dropped tx frames when transmitted fast.
2. implemented 'cntrl device' for Windows in wanpipe_tdm_api.c
*/